/*************************************************************************
 *
 *    Used with ICCARM and AARM.
 *
 *
 *    File name   : EMC_drv.c
 *    Description : SDRAM driver K4S281632F 128Mb
 *                  and NAND Flash HY27UF081G2M driver
 *    History :
 *    1. Date        : 6, March 2008
 *       Author      : Stanimir Bonev
 *       Description : Create
 *    2. Date        : June 2008
 *       Author      : Stoyan Choynev
 *       Description : SDRAM Driver ported HY57V561620FTP-H device
 *                     added NAND Flash Driver
 *    3. Date        : 12 May 2010
 *       Author      : Aleksandar Mitev
 *       Description : NAND driver now supports the HY27UF081G2M chip. Also made some minor
 *				   changes to the SDRAM driver to support K4S281632F by Samsung
 *
 **************************************************************************/

#include "includes.h"

#if   SYS_FREQ == (72MHZ)
#define EMC_PERIOD          13.8  // 72MHz
#elif   SYS_FREQ == (57MHZ)
#define EMC_PERIOD          17.4  // 57.6MHz
#elif SYS_FREQ == (48MHZ)
#define EMC_PERIOD          20.8  // 48MHz
#elif SYS_FREQ == (36MHZ)
#define EMC_PERIOD          27.8  // 36MHz
#elif SYS_FREQ == (24MHZ)
#define EMC_PERIOD          41.7  // 24MHz
#else
#error Frequency not defined
#endif

#define P2C(Period)           (((Period<EMC_PERIOD)?0:(Int32U)((Flo32)Period/EMC_PERIOD))+1)

/*************************************************************************
* SDRAM definitions
**************************************************************************/
#define SDRAM_BASE_ADDR *(volatile Int32U*)0xA0000000
#define SDRAM_SIZE 0x01000000 // 16MB

#define SDRAM_REFRESH         7813
#define SDRAM_TRP             20
#define SDRAM_TRAS            42
#define SDRAM_TAPR            1
#define SDRAM_TDAL            2
#define SDRAM_TWR             2
#define SDRAM_TRC             63
#define SDRAM_TRFC            63
#define SDRAM_TXSR            1
#define SDRAM_TRRD            15
#define SDRAM_TMRD            2

/*************************************************************************
* NAND Flash
**************************************************************************/
#define NAND_BASE_ADDR  *(volatile Int8U*)0x81000000
#define NAND_CMD_LATCH  *(volatile Int8U*)0x81040000
#define NAND_ADDR_LATCH *(volatile Int8U*)0x81080000

#define NAND_tCSWE            0
#define NAND_tCSOW            0
/*#define NAND_tREA             100//60
#define NAND_tRC              200//120
#define NAND_tWP              100//60
#define NAND_tTURNE           200//150
*/
#define NAND_tREA             60
#define NAND_tRC              120
#define NAND_tWP              60
#define NAND_tTURNE           150

#define NAND_W_TO_B_DLY       20       //

#define NAND_MAIN_SIZE      		 (2048)
#define NAND_SPARE_SIZE     		 (64)
#define NAND_PAGE_SIZE       		 (NAND_MAIN_SIZE+NAND_SPARE_SIZE)
#define NAND_BLK_NUM         		 (1024)
#define NAND_PG_PER_BLK      		 (64)
#define NAND_PG_NUM           	 (NAND_BLK_NUM*NAND_PG_PER_BLK)
#define NAND_INVALID_FLAG_OFFSET    (NAND_MAIN_SIZE + 0)


#define NAND_PASS (1<<0)
#define NAND_WP   (1<<7)

#define NAND_RB_PIN (1<<12) // P2.12

// these defines are empty since CS1 is hardware implemented by the parallel port module
#define NandChipEn()
#define NandChipDis()


//Wait until NAND gets ready
void NandWaitReady(void)
{
  for(int i = 0; i < NAND_W_TO_B_DLY; i++);

  while(!(FIO2PIN & NAND_RB_PIN));
}

//Read device status
char NandGetStatus(void)
{
  NAND_CMD_LATCH = 0x70;

  NandWaitReady();

  return NAND_BASE_ADDR;
}

/*************************************************************************
 * Function Name: NandReset
 * Parameters: None
 *
 * Return: None
 *
 * Description:Reset the NAND Flash
 *
 *************************************************************************/
void NandReset(void)
{
  NandChipEn();

  NAND_CMD_LATCH = 0xFF;

  NandWaitReady();

  NandChipDis();
}

/*************************************************************************
 * Function Name: NandGetID
 * Parameters: pID - Pointer to char array. The array should be at least
 *                   4 bytes long
 *
 * Return: None
 *
 * Description:Reads the Device ID.
 *
 *************************************************************************/
void NandGetID(char * pID)
{
  NandChipEn();
  //ID Command
  NAND_CMD_LATCH = 0x90;
  //
  NAND_ADDR_LATCH = 0x00;

  NandWaitReady();

  for(int i = 0 ; i < 4; i++) *pID++ = NAND_BASE_ADDR;

  NandChipDis();
}

/*************************************************************************
 * Function Name: NandRead
 * Parameters: pDest - pointer to char buffer.
 *             PageIndex - Index to the Page
 *             Column - Column Index
 *                      from 0 to 255 Main Array 1st Half
 *                      from 256 to 511 Main Array 2nd Half
 *                      from 511 to 527 SPare Array
 *             Size - Bytes to be read
 *
 * Return: none
 *
 * Description: Reads one page from NAND Flash
 *
 *************************************************************************/
void NandRead(char * pDest,Int32U PageIndex, Int32U Column, Int32U Size)
{
  assert(NAND_PG_NUM > PageIndex);
  assert(NAND_PAGE_SIZE >= (Size+Column));

  NandChipEn();

	// issue command for reading
	NAND_CMD_LATCH = 0x00;
	//column address
	NAND_ADDR_LATCH = (Column >> 0) & 0xFF;
	NAND_ADDR_LATCH = (Column >> 8) & 0x0F;
	//Page Address
	NAND_ADDR_LATCH = (PageIndex>>0) & 0xFF;
	NAND_ADDR_LATCH = (PageIndex>>8) & 0xFF;

	NAND_CMD_LATCH = 0x30;
	
	NandWaitReady();
	
	for(Int32U cnt=0; Size > cnt; cnt++)
		*pDest++ = NAND_BASE_ADDR;
	
	NandChipDis();
}
/*************************************************************************
 * Function Name: NandReadPage
 * Parameters: pDest - pointer to char buffer. The Buffer's size should be at least
 *                     NAND_PAGE_SIZE.
 *             Page - Index to the Page
 * Return: none
 *
 * Description: Reads one page from NAND Flash
 *
 *************************************************************************/
void NandReadPage(char * pDest,Int32U Page)
{
  assert(NAND_PG_NUM > Page);

  NandChipEn();
  //Read 1 First page half
  NAND_CMD_LATCH = 0x00;
  //column 0
  NAND_ADDR_LATCH = 0;
  //Page Address
  NAND_ADDR_LATCH = (Page>>0) & 0xFF;
  NAND_ADDR_LATCH = (Page>>8) & 0xFF;
  NAND_ADDR_LATCH = (Page>>16)& 0x01;

  NandWaitReady();

  for(Int32U cnt=0; NAND_PAGE_SIZE > cnt; cnt++) *pDest++ = NAND_BASE_ADDR;

  NandChipDis();
}

/*************************************************************************
 * Function Name: NandSeqRead
 * Parameters: pDest - pointer to char buffer. The Buffer's size should be at least
 *                     NAND_PG_SIZE * NAND_PGS_PER_BLK.
 *             PageIndex - Index to the Page
 * Return: none
 *
 * Description: Reads All pages to the end of the Block
 *
 *************************************************************************/
void NandSeqRead(Int8U * pDest,Int32U PageIndex)
{
  assert(NAND_PG_NUM > PageIndex);

  NandChipEn();

  do
  {
	//Read1
	NAND_CMD_LATCH = 0x00;
	//column 0
	NAND_ADDR_LATCH = 0;
	NAND_ADDR_LATCH = 0;
	//Page Address
	NAND_ADDR_LATCH = (PageIndex>>0) & 0xFF;
	NAND_ADDR_LATCH = (PageIndex>>8) & 0xFF;
	
	NAND_CMD_LATCH = 0x30;
	
	NandWaitReady();

    for(Int32U cnt=0; NAND_PAGE_SIZE > cnt; cnt++) *pDest++ = NAND_BASE_ADDR;

  }while((++PageIndex) & (NAND_PG_PER_BLK-1));

  NandChipDis();
}

/*************************************************************************
 * Function Name: NandProgramPage
 * Parameters: pSrc - pointer to char buffer.
 *
 *             PageIndex - Index to the Page
 * Return: Boolean TRUE - Write O.K.
 *                 FALSE - WRITE Error
 * Description: Writes one page to NAND Flash
 *
 *************************************************************************/
Int8U NandProgramPage(Int8U * pSrc,Int32U Page)
{
Int8U NandStatus;
  assert(NAND_PG_NUM > Page);

  NandChipEn();

  NAND_CMD_LATCH = 0x80;
  //column 0
  NAND_ADDR_LATCH = 0;
  NAND_ADDR_LATCH = 0;
  //Page Address
  NAND_ADDR_LATCH = (Page>>0) & 0xFF;
  NAND_ADDR_LATCH = (Page>>8) & 0xFF;

  NandWaitReady();

  for( Int32U cnt=0; NAND_PAGE_SIZE > cnt; cnt++) NAND_BASE_ADDR = *pSrc++;

  NAND_CMD_LATCH = 0x10;

  NandWaitReady();

  NandStatus = NandGetStatus();

  NandChipDis();

  return NandStatus;
}


/*************************************************************************
 * Function Name: NandEraseBlock
 * Parameters: Block.
 *
 *
 * Return: Boolean TRUE - Erase O.K.
 *                 FALSE - Erase Error
 * Description: Erase a block.
 *
 *************************************************************************/
Int8U NandEraseBlock(Int32U Block)
{
Int32U Page = Block * NAND_PG_PER_BLK;
Int8U NandStatus;

  assert(NAND_BLK_NUM > Block);

  NandChipEn();

  NAND_CMD_LATCH = 0x60;

  NAND_ADDR_LATCH = (Page>>0) & 0xFF;
  NAND_ADDR_LATCH = (Page>>8) & 0xFF;

  NAND_CMD_LATCH = 0xD0;

  NandWaitReady();

  NandStatus = NandGetStatus();

  NandChipDis();

  return NandStatus;
}

Boolean NAND_Test(void);
Boolean NAND_TestBlock(int Block);

/*************************************************************************
 * Function Name: NandCheckBlock
 * Parameters: Block - block index to test
 *
 * Return: TRUE if block is valid, FALSE if block is bad
 *
 * Description: Checks a byte in the spare area, which indicates if the byte is bad
 * 			 For the HY27UF081G2M device: Any block where the 1st Byte in the spare area of
 *			 the 1st or 2nd page(if the 1st page is Bad) does not contain FFh is a Bad Block.
 *
 *************************************************************************/
Boolean NandCheckBlock(Int32U Block)
{
	Int32U Page;
	char Byte;

	//For the first and second page in the Block
	for(Page = 0; 2 > Page; Page++)
	{
		//the 1th Byte in the Spare array shoud be 0xFF
		NandRead(&Byte , Block * NAND_PG_PER_BLK + Page, NAND_INVALID_FLAG_OFFSET, 1);
		if(0xFF != Byte)
			return FALSE;
	}
	return (TRUE);
}
/*************************************************************************
 * Function Name: EMC_Init
 * Parameters: none
 *
 * Return: none
 *
 * Description: SDRAM controller and memory init
 *
 *************************************************************************/
void EMC_Init(void)
{
  // Assign pins to EMC controller(SDRAM & NAND Flash)
	
  // configure P2.12 - R/#B on NAND Flash
  PINSEL4 &= BIN32(11111100,11111111,11111111,11111111);
  PINMODE4&= BIN32(11111100,11111111,11111111,11111111);

  // configure #RAS, #CAS, CLKOUT0, #DYCS0, CKEOUT0, DQMOUT0, DQMOUT1
  PINSEL5 &= BIN32(11110000,11111100,11111100,11000000);
  PINSEL5 |= BIN32(00000101,00000001,00000001,00010101);
  PINMODE5&= BIN32(11110000,11111100,11111100,11000000);
  PINMODE5|= BIN32(00001010,00000010,00000010,00101010);

  // configure D0 - D15
  PINSEL6  = BIN32(01010101,01010101,01010101,01010101);
  PINMODE6 = BIN32(10101010,10101010,10101010,10101010);

  // configure A0 - A14
  PINSEL8 &= BIN32(11000000,00000000,00000000,00000000);
  PINSEL8 |= BIN32(00010101,01010101,01010101,01010101);
  PINMODE8&= BIN32(11000000,00000000,00000000,00000000);
  PINMODE8|= BIN32(00101010,10101010,10101010,10101010);

  // configure A18, A19, #OE, #WE, #CS1
  PINSEL9 &= BIN32(00111111,11110000,11111111,00001111);
  PINSEL9 |= BIN32(01000000,00000101,00000000,01010000);
  PINMODE9&= BIN32(00111111,11110000,11111111,11111111);
  PINMODE9|= BIN32(10000000,00001010,00000000,10100000);

  NandChipDis();

  // Init SDRAM controller
  // Enable EMC clock
  PCONP_bit.PCEMC = 1;
  EMCCONTROL      = 1;         // enable EMC

  EMCDINAMICRDCFG = 1;
  EMCDYNAMICRASCAS0_bit.CAS = 3;
  EMCDYNAMICRASCAS0_bit.RAS = 3;
  EMCDYNAMICRP = P2C(SDRAM_TRP);
  EMCDYNAMICRAS = P2C(SDRAM_TRAS);
  EMCDYNAMICSREX = SDRAM_TXSR;
  EMCDYNAMICAPR = SDRAM_TAPR;
  EMCDYNAMICDAL = SDRAM_TDAL+P2C(SDRAM_TRP);
  EMCDYNAMICWR = SDRAM_TWR;
  EMCDYNAMICRC = P2C(SDRAM_TRC);
  EMCDYNAMICRFC = P2C(SDRAM_TRFC);
  EMCDYNAMICXSR = SDRAM_TXSR;
  EMCDYNAMICRRD = P2C(SDRAM_TRRD);
  EMCDYNAMICMRD = SDRAM_TMRD;
  EMCDYNAMICCFG0 = 0x0000480;        // 12 row, 9 - col, SDRAM
  // JEDEC General SDRAM Initialization Sequence
  // DELAY to allow power and clocks to stabilize ~100 us
  // NOP
  EMCDINAMICCTRL = 0x0183;
  for(volatile Int32U i = 200*30; i;i--);
  // PALL
  EMCDINAMICCTRL_bit.I = 2;
  EMCDINAMICRFR = 1;
  for(volatile Int32U i= 128; i; --i); // > 128 clk
  EMCDINAMICRFR = P2C(SDRAM_REFRESH) >> 4;
  // COMM
  EMCDINAMICCTRL_bit.I = 1;
  // Burst 8, Sequential, CAS-2
  volatile unsigned long Dummy = *(volatile unsigned short *)
                                  ((Int32U)&SDRAM_BASE_ADDR + (0x33UL << (12)));
  // NORM
  EMCDINAMICCTRL = 0x0000;
  EMCDYNAMICCFG0_bit.B = 1;
  for(volatile Int32U i = 10000; i;i--);
  //8Bit Bus
  EMCSTATICCNFG1_bit.MW = 0;
  //Disable Page mode
  EMCSTATICCNFG1_bit.PM = 0;
  //Chip select active low
  EMCSTATICCNFG1_bit.PC = 0;
  //No Extended wait
  EMCSTATICCNFG1_bit.EW = 0;
  //Enable byte line state bit
  EMCSTATICCNFG1_bit.PB = 1;
  EMCSTATICCNFG1_bit.B = 0;
  EMCSTATICCNFG1_bit.P = 0;
  //Set delays
  EMCSTATICWAITWEN1 = NAND_tCSWE;
  EMCSTATICWAITOEN1 = NAND_tCSOW;
  EMCSTATICWAITRD1 = P2C(NAND_tREA);
  EMCSTATICWAITPG1 = 0;
  EMCSTATICWAITWR1 = P2C(NAND_tWP);
  //Reset
  NandReset();
}

/*************************************************************************
 * Function Name: SDRAM_Test
 * Parameters: none
 *
 * Return: TRUE if test passed, FALSE if an error occured during test
 *
 * Description: Performes a test of the SDRAM
 *
 *************************************************************************/
Boolean SDRAM_Test(void)
{
	volatile Int16U tmp;
	volatile Int16U dummy;
	
	printf("SDRAM Test Start\n\r");
	
	/* TEST OF DATA BUS */
#define OFFSET_FOR_TEST_DATA_BUS 0x04
	
	printf("Data bus test start\n\r");
	
	// test all 1s
	*(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS) = (Int16U)0xFFFFFFFF;
	/*flush read buffers!*/
	for(volatile int k = 0; (16*16)/sizeof(Int16U) > k; k++) {
		dummy = *(((Int16U*)&SDRAM_BASE_ADDR)+SDRAM_SIZE/4+k);
	}
	if ((tmp = *(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS)) != (Int16U)0xFFFFFFFF) {
		printf("SDRAM Data bus error, written 0x%X, read 0x%X\n\r", (Int16U)0xFFFFFFFF, tmp);
		return(FALSE);
	}
	
	// test all 0s
	*(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS) = (Int16U)0x00000000;
	/*flush read buffers!*/
	for(volatile int k = 0; (16*16)/sizeof(Int16U) > k; k++) {
		dummy = *(((Int16U*)&SDRAM_BASE_ADDR)+SDRAM_SIZE/4+k);
	}
	if ((tmp = *(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS)) != (Int16U)0x00000000) {
		printf("SDRAM Data bus error, written 0x%X, read 0x%X\n\r", (Int16U)0x00000000, tmp);
		return(FALSE);
	}
	
	// test moving 1
	for(int i = 0; i < (sizeof(Int16U) * 8); i++) {
		*(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS) = (Int16U)(0x1 << i);
		/*flush read buffers!*/
		for(volatile int k = 0; (16*16)/sizeof(Int16U) > k; k++) {
			dummy = *(((Int16U*)&SDRAM_BASE_ADDR)+SDRAM_SIZE/4+k);
		}
		if ((tmp = *(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS)) != (Int16U)(0x1 << i)) {
			printf("SDRAM Data bus error, on bit %d\n\r", i);
			return(FALSE);
		}
	}
	
	// test moving 0
	for(int i = 0; i < (sizeof(Int16U) * 8); i++) {
		*(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS) = (Int16U)(0xFFFFFFFF & ~(0x1 << i));
		/*flush read buffers!*/
		for(volatile int k = 0; (16*16)/sizeof(Int16U) > k; k++) {
			dummy = *(((Int16U*)&SDRAM_BASE_ADDR)+SDRAM_SIZE/4+k);
		}
		if ((tmp = *(Int16U*)((Int32U)&SDRAM_BASE_ADDR+OFFSET_FOR_TEST_DATA_BUS)) != (Int16U)(0xFFFFFFFF & ~(0x1 << i))) {
			printf("SDRAM Data bus error, on bit %d\n\r", i);
			return(FALSE);
		}
	}
	printf("Data bus test end - SUCCESS\n\r");
	
	
	
  // 32 bits access
  for (Int32U i = 0; i < SDRAM_SIZE; i+=sizeof(Int32U))
  {
    *(Int32U*)((Int32U)&SDRAM_BASE_ADDR+i) = i;
  }

  for (Int32U i = 0; i < SDRAM_SIZE; i+=sizeof(Int32U))
  {
    if (*(Int32U*)((Int32U)&SDRAM_BASE_ADDR+i) != i)
    {
      printf("SDRAM Verification (32bit) error on address : 0x%x\n\r",(Int32U)&SDRAM_BASE_ADDR+i);
      return(FALSE);
    }
  }

  // 16 bits access
  for (Int32U i = 0; i < 0x10000; i+=sizeof(Int16U))
  {
    *(Int16U*)((Int32U)&SDRAM_BASE_ADDR+i) = i;
  }

  for (Int32U i = 0; i < 0x10000; i+=sizeof(Int16U))
  {
    if (*(Int16U*)((Int32U)&SDRAM_BASE_ADDR+i) != i)
    {
      printf("SDRAM Verification (16bit) error on address : 0x%x\n\r",(Int32U)&SDRAM_BASE_ADDR+i);
      return(FALSE);
    }
  }

  // 8 bits access
  for (Int32U i = 0; i < 0x100; i+=sizeof(Int8U))
  {
    *(Int8U*)((Int32U)&SDRAM_BASE_ADDR+i) = i;
  }

  for (Int32U i = 0; i < 0x100; i+=sizeof(Int8U))
  {
    if (*(Int8U*)((Int32U)&SDRAM_BASE_ADDR+i) != i)
    {
      printf("SDRAM Verification (8bit) error on address : 0x%x\n\r",(Int32U)&SDRAM_BASE_ADDR+i);
      return(FALSE);
    }
  }
  printf("SDRAM Test Pass\n\r");
  return(TRUE);
}

/*************************************************************************
 * Function Name: NAND_Test
 * Parameters: none
 *
 * Return: TRUE if test passed, FALSE if an error occured during test
 *
 * Description: Performes a test on the whole NAND flash
 *
 *************************************************************************/
Boolean NAND_Test(void)
{
	char idBuf[4];
	Int32U Block;

  printf("NAND Test Start\n\r");

  NandGetID(idBuf);

  if((0xAD != idBuf[0]) || (0xF1 != idBuf[1]) )
  {
    printf("NAND Flash not supported or not present\n\r");
    printf("Bytes read: %#x, %#x, %#x, %#x\n\r", idBuf[0], idBuf[1], idBuf[2], idBuf[3]);
    return (FALSE);
  }
  printf("Found device ID: %#x, %#x, %#x, %#x\n\r", idBuf[0], idBuf[1], idBuf[2], idBuf[3]);

  for(Block = 0; NAND_BLK_NUM > Block; Block++)
  {
	
	NAND_TestBlock(Block);
  }

  printf("NAND Test End\n\r");
  return (TRUE);
}


/*************************************************************************
 * Function Name: NAND_TestBlock
 * Parameters: none
 *
 * Return: TRUE if test passed, FALSE if an error occured during test
 *
 * Description: Performes a test on a specific block in the NAND flash
 * *NOTE: due to the large size of the block of the flash, SDRAM address space
 * 	is used as a RAM storage for the test!
 *
 *************************************************************************/
Boolean NAND_TestBlock(int Block)
{
	Int8U OperationStatus;
	Boolean Error = FALSE;
	Int8U * pBuffer;

	//Get Buffer for NAND Read/Write
	//pBuffer = malloc(NAND_PG_PER_BLK*NAND_PAGE_SIZE);
	pBuffer = (Int8U*)((Int32U)&SDRAM_BASE_ADDR);
	
	assert(pBuffer);

    //Check for invalid Block
    if((FALSE) == NandCheckBlock(Block))
    {//Do not erase invalid Block
      printf("Found invalid Block: %u\n\r", Block);
      return FALSE;
    }

    OperationStatus = NandEraseBlock(Block);

    if( !(NAND_WP & OperationStatus) )
    {
      printf("NAND is Write Protected.\n\r");
      free(pBuffer);
      return(FALSE);
    }

    if(NAND_PASS & OperationStatus)
    {
      printf("NAND Erase Failed Block: %u\n\r", Block);
      return FALSE;
    }

// make 0 if you just want to erase the NAND (leaving bad blocks untouched)
#if 1
    for(Int32U Page = 0; NAND_PG_PER_BLK > Page; Page++)
    {
      for(Int32U i = 0; NAND_PAGE_SIZE > i; i++)
      {
        pBuffer[i] = ((Block^Page) + i) & 0xFF;
        //Valid Block Mark
        if(((0 == Page) || (1 == Page)) && NAND_INVALID_FLAG_OFFSET == i )
	        pBuffer[i] = 0xFF;
      }

      OperationStatus = NandProgramPage(pBuffer, Block*NAND_PG_PER_BLK + Page);

      if(!(NAND_WP & OperationStatus))
      {
        printf("NAND is Write Protected\n\r");
        free(pBuffer);
        return(FALSE);
      }

      if(NAND_PASS & OperationStatus)
      {
        printf("NAND Program Error Page: %u Block: %u\n\r", Page, Block);
        Error= TRUE;
        break;
      }
    }
    if((FALSE) == Error)
    {//Read whole Block
      NandSeqRead(pBuffer,Block*NAND_PG_PER_BLK);

      for(Int32U Page = 0; NAND_PG_PER_BLK > Page; Page++)
      {
        for(Int32U i = 0; NAND_PAGE_SIZE > i; i++)
        {
        char Test = ((Block^Page) + i) & 0xFF;
          if(((0 == Page) || (1 == Page)) && NAND_INVALID_FLAG_OFFSET == i ) Test = 0xFF;

          if( pBuffer[Page*NAND_PAGE_SIZE+i] != Test)
          {
            printf("NAND Read Error Page %u Block %u\n\r", Page, Block);
            Error = TRUE;
            break;
          }
        }
      }
    }

    if((FALSE) == Error)
    {
      printf("NAND Block: %u O.K.\n\r", Block);
      return TRUE;
    }
#endif

    return FALSE;
}